using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;
using System.IO;

namespace GbaSPCBuilder
{
	/// <summary>
	/// Summary description for Form1.
	/// </summary>
	public class SPCRomBuilder : System.Windows.Forms.Form
	{
		private Button addFiles;
		private ListView fileList;
		private System.Windows.Forms.ColumnHeader spcFilename;
		private System.Windows.Forms.ColumnHeader spcSize;
		/// <summary>
		/// Required designer variable.
		/// </summary>
		private System.ComponentModel.Container components = null;

		private int headerPos = -1;
		private string romName = "gbaspc.gba";

		public SPCRomBuilder()
		{
			InitializeComponent();

			this.fileList.View = View.Details;
			this.fileList.FullRowSelect = true;
		}

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if (components != null) 
				{
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			this.addFiles = new System.Windows.Forms.Button();
			this.fileList = new System.Windows.Forms.ListView();
			this.spcFilename = new System.Windows.Forms.ColumnHeader();
			this.spcSize = new System.Windows.Forms.ColumnHeader();
			this.SuspendLayout();
			// 
			// addFiles
			// 
			this.addFiles.Location = new System.Drawing.Point(8, 224);
			this.addFiles.Name = "addFiles";
			this.addFiles.TabIndex = 0;
			this.addFiles.Text = "Add Files";
			this.addFiles.Click += new System.EventHandler(this.addFiles_Click);
			// 
			// fileList
			// 
			this.fileList.Columns.AddRange(new System.Windows.Forms.ColumnHeader[] {
																					   this.spcFilename,
																					   this.spcSize});
			this.fileList.Location = new System.Drawing.Point(8, 8);
			this.fileList.Name = "fileList";
			this.fileList.Size = new System.Drawing.Size(392, 208);
			this.fileList.TabIndex = 1;
			// 
			// spcFilename
			// 
			this.spcFilename.Text = "Filename";
			this.spcFilename.Width = 250;
			// 
			// spcSize
			// 
			this.spcSize.Text = "Size";
			this.spcSize.Width = 100;
			// 
			// SPCRomBuilder
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 13);
			this.ClientSize = new System.Drawing.Size(408, 266);
			this.Controls.Add(this.fileList);
			this.Controls.Add(this.addFiles);
			this.Name = "SPCRomBuilder";
			this.Text = "SPCRomBuilder";
			this.Load += new System.EventHandler(this.Form1_Load);
			this.ResumeLayout(false);

		}
		#endregion

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main(string[] args) 
		{
			SPCRomBuilder builder = new SPCRomBuilder();
			if (args.Length == 0)
			{
				Application.Run(builder);
			} 
			else
			{
				if (args[0] == "-add")
				{
					if (args.Length < 2)
					{
						Console.WriteLine("Need filename");
						return;
					}

					builder.Initialize();
					for (int i=1; i<args.Length; i++)
					{
						DirectoryInfo directory = new DirectoryInfo(Environment.CurrentDirectory);
						FileInfo[] files = directory.GetFiles(args[i]);
						foreach (FileInfo fileInfo in files)
						{
							Console.WriteLine("Adding: " + fileInfo.Name);
							builder.AddFileToEnd(fileInfo.FullName);
						}
					}
				}
			}
		}

		public void AddFileToEnd(string filename)
		{
			FileInfo fileInfo = new FileInfo(filename);
			FileStream stream = fileInfo.OpenRead();

			// Tack on the SPC file
			FileStream rom = new FileStream(romName, FileMode.Append);
			int startPos = (int)rom.Length;

			string tmpname = fileInfo.Name;
			if (tmpname.Length - 4 >= 0 && tmpname[tmpname.Length - 4] == '.')
				tmpname = tmpname.Substring(0, tmpname.Length - 4);

			for (int i=0; i<64; i++)
			{
				if (i >= tmpname.Length)
					rom.WriteByte(0);
				else
					rom.WriteByte((byte)(tmpname[i]));
			}

			byte[] spcdata = new byte[stream.Length];
			stream.Read(spcdata, 0, (int)stream.Length);
			rom.Write(spcdata, 0, (int)stream.Length);
			rom.Close();

			// Seek back into the table and update the position
			rom = new FileStream(romName, FileMode.Open);
			BinaryReader binaryReader = new BinaryReader(rom);
			BinaryWriter binaryWriter = new BinaryWriter(rom);
			for (int i=0; i<1024; i++)
			{
				rom.Seek(this.headerPos + 8 + (i * 4), SeekOrigin.Begin);
				uint offset = binaryReader.ReadUInt32();
				if (offset == 0)
				{
					rom.Seek(this.headerPos + 8 + (i * 4), SeekOrigin.Begin);
					binaryWriter.Write((uint)startPos);
					break;
				}
			}
			rom.Close();

			stream.Close();
		}

		private void addFiles_Click(object sender, System.EventArgs e)
		{
			OpenFileDialog fileDialog = new OpenFileDialog();
			fileDialog.Filter = "SPC files (*.spc)|*.spc|All files (*.*)|*.*";
			fileDialog.Multiselect = true;

			if (fileDialog.ShowDialog(this) == DialogResult.OK)
			{
				foreach (string filename in fileDialog.FileNames)
				{
					this.AddFileToEnd(filename);
				}
			}	
	
			this.DisplayFiles();
		}

		private int ParseRom(string romName)
		{
			byte[] header = new byte[]{123,154,143,36,201,243,9,255};
			try 
			{
				FileStream rom = new FileStream(romName, FileMode.Open);
				int length = (int)rom.Length;

				int headerPos = -1;
				for (int i = 0; i < length; i += 256)
				{
					if (i + 8 > length) break;
					bool good = true;
					rom.Seek(i, SeekOrigin.Begin);
					for (int j = 0; j < 8; j++)
					{
						if (rom.ReadByte() != header[j])
							good = false;
					}
					if (good)
						headerPos = i;
				}
				rom.Close();

				if (headerPos == -1)
				{
					// No header found, so pad the rom up to nearest 256 bytes and add
					// header information.
					int numAppend = length % 256;
					if (numAppend != 0) numAppend = 256 - numAppend;

					rom = new FileStream(romName, FileMode.Append);
					for (int i = 0; i < numAppend; i++) rom.WriteByte(0);
					headerPos = (int)rom.Length;
					for (int i = 0; i < 8; i++) rom.WriteByte(header[i]);
					for (int i = 0; i < 1024; i++)
					{
						// Dword represeting offset
						rom.WriteByte(0);
						rom.WriteByte(0);
						rom.WriteByte(0);
						rom.WriteByte(0);
					}
					rom.Close();
				}

				return headerPos;
			}
			catch (FileNotFoundException)
			{
				MessageBox.Show(this, "Unable to locate gbaspc.gba.  It needs to be in the same directory as this program");
				this.Close();
			}

			return -1;
		}

		private void DisplayFiles()
		{
			this.fileList.Items.Clear();

			FileStream rawStream = new FileStream(romName, FileMode.Open);
			BinaryReader rom = new BinaryReader(rawStream);

			// Parse all the filenames
			for (int i=0; i<1024; i++)
			{
				rom.BaseStream.Seek(this.headerPos + 8 + (i * 4), SeekOrigin.Begin);
				uint offset = rom.ReadUInt32();

				if (offset != 0)
				{
					rom.BaseStream.Seek(offset, SeekOrigin.Begin);
					string filename = "";
					for (int j=0; j<64; j++)
					{
						int c = rom.BaseStream.ReadByte();
						if (c == 0) break;
						filename += (char)c;
					}
					string[] items = new string[]{filename,"-"};
					ListViewItem listItem = new ListViewItem(items);
					this.fileList.Items.Add(listItem);
				} else break;
			}

			rom.Close();
		}

		public void Initialize()
		{
			this.romName = Environment.CurrentDirectory + "\\" + this.romName;
//			FileStream rom = new FileStream(romName, FileMode.OpenOrCreate);
//			rom.Close();
			this.headerPos = this.ParseRom(romName);
			if (this.headerPos != -1)
			{
				this.DisplayFiles();
			} 
			else
			{
				string[] items = new string[]{"No header found...",""};
				ListViewItem listItem = new ListViewItem(items);
				this.fileList.Items.Add(listItem);
			}
		}

		private void Form1_Load(object sender, System.EventArgs e)
		{
			this.Initialize();
		}
	}
}
